from typing import Iterable, Optional, Tuple, Type, TypeVar

import numpy as np
import pytest
import scipy  # pyright: ignore[reportMissingTypeStubs]
import xarray as xr
from typing_extensions import TypeGuard

from bioimageio.core.axis import AxisId
from bioimageio.core.common import MemberId
from bioimageio.core.sample import Sample
from bioimageio.core.stat_calculators import compute_measures
from bioimageio.core.stat_measures import SampleMean, SampleQuantile, SampleStd
from bioimageio.core.tensor import Tensor


@pytest.fixture(scope="module")
def tid():
    return MemberId("data123")


def test_scale_linear(tid: MemberId):
    from bioimageio.core.proc_ops import ScaleLinear

    offset = xr.DataArray([1, 2, 42], dims=("channel",))
    gain = xr.DataArray([1, 2, 3], dims=("channel",))
    data = xr.DataArray(np.arange(6).reshape((1, 2, 3)), dims=("x", "y", "channel"))
    sample = Sample(members={tid: Tensor.from_xarray(data)}, stat={}, id=None)

    op = ScaleLinear(input=tid, output=tid, offset=offset, gain=gain)
    op(sample)

    expected = xr.DataArray(
        np.array([[[1, 4, 48], [4, 10, 57]]]), dims=("x", "y", "channel")
    )
    xr.testing.assert_allclose(expected, sample.members[tid].data, rtol=1e-5, atol=1e-7)


def test_scale_linear_no_channel(tid: MemberId):
    from bioimageio.core.proc_ops import ScaleLinear

    op = ScaleLinear(tid, tid, offset=1, gain=2)
    data = xr.DataArray(np.arange(6).reshape(2, 3), dims=("x", "y"))
    sample = Sample(members={tid: Tensor.from_xarray(data)}, stat={}, id=None)
    op(sample)

    expected = xr.DataArray(np.array([[1, 3, 5], [7, 9, 11]]), dims=("x", "y"))
    xr.testing.assert_allclose(expected, sample.members[tid].data, rtol=1e-5, atol=1e-7)


T = TypeVar("T")


def is_iterable(val: Iterable[T], inner: Type[T]) -> TypeGuard[Iterable[T]]:
    """Determines whether all objects in the list are strings"""
    return all(isinstance(x, inner) for x in val)


def test_zero_mean_unit_variance(tid: MemberId):
    from bioimageio.core.proc_ops import ZeroMeanUnitVariance

    data = xr.DataArray(np.arange(9).reshape(3, 3), dims=("x", "y"))
    sample = Sample(members={tid: Tensor.from_xarray(data)}, stat={}, id=None)
    m = SampleMean(member_id=tid)
    std = SampleStd(member_id=tid)
    op = ZeroMeanUnitVariance(tid, tid, m, std)
    req = op.required_measures
    sample.stat = compute_measures(req, [sample])
    op(sample)

    expected = xr.DataArray(
        np.array(
            [
                [-1.54919274, -1.16189455, -0.77459637],
                [-0.38729818, 0.0, 0.38729818],
                [0.77459637, 1.16189455, 1.54919274],
            ]
        ),
        dims=("x", "y"),
    )
    xr.testing.assert_allclose(expected, sample.members[tid].data, rtol=1e-5, atol=1e-7)


def test_zero_mean_unit_variance_fixed(tid: MemberId):
    from bioimageio.core.proc_ops import FixedZeroMeanUnitVariance

    op = FixedZeroMeanUnitVariance(
        tid,
        tid,
        mean=xr.DataArray([3, 4, 5], dims=("channel",)),
        std=xr.DataArray([2.44948974, 2.44948974, 2.44948974], dims=("channel",)),
    )
    data = xr.DataArray(np.arange(9).reshape((1, 3, 3)), dims=("batch", "channel", "x"))
    expected = xr.DataArray(
        np.array(
            [
                [
                    [-1.22474487, -0.81649658, -0.40824829],
                    [-0.40824829, 0.0, 0.40824829],
                    [0.40824829, 0.81649658, 1.22474487],
                ]
            ]
        ),
        dims=("batch", "channel", "x"),
    )
    sample = Sample(members={tid: Tensor.from_xarray(data)}, stat={}, id=None)
    op(sample)
    xr.testing.assert_allclose(expected, sample.members[tid].data, rtol=1e-5, atol=1e-7)


def test_zero_mean_unit_variance_fixed2(tid: MemberId):
    from bioimageio.core.proc_ops import FixedZeroMeanUnitVariance

    np_data = np.arange(9).reshape(3, 3)
    mean = float(np_data.mean())
    std = float(np_data.mean())
    eps = 1.0e-7
    op = FixedZeroMeanUnitVariance(tid, tid, mean=mean, std=std, eps=eps)

    data = xr.DataArray(np_data, dims=("x", "y"))
    sample = Sample(members={tid: Tensor.from_xarray(data)}, stat={}, id=None)
    expected = xr.DataArray((np_data - mean) / (std + eps), dims=("x", "y"))
    op(sample)
    xr.testing.assert_allclose(expected, sample.members[tid].data, rtol=1e-5, atol=1e-7)


def test_zero_mean_unit_across_axes(tid: MemberId):
    from bioimageio.core.proc_ops import ZeroMeanUnitVariance

    data = xr.DataArray(np.arange(18).reshape((2, 3, 3)), dims=("channel", "x", "y"))

    op = ZeroMeanUnitVariance(
        tid,
        tid,
        SampleMean(member_id=tid, axes=(AxisId("x"), AxisId("y"))),
        SampleStd(member_id=tid, axes=(AxisId("x"), AxisId("y"))),
    )
    sample = Sample(members={tid: Tensor.from_xarray(data)}, stat={}, id=None)
    sample.stat = compute_measures(op.required_measures, [sample])

    expected = xr.concat(
        [(data[i : i + 1] - data[i].mean()) / data[i].std() for i in range(2)],
        dim="channel",
    )
    op(sample)
    xr.testing.assert_allclose(expected, sample.members[tid].data, rtol=1e-5, atol=1e-7)


def test_binarize(tid: MemberId):
    from bioimageio.core.proc_ops import Binarize

    op = Binarize(tid, tid, threshold=14)
    data = xr.DataArray(np.arange(30).reshape((2, 3, 5)), dims=("x", "y", "channel"))
    sample = Sample(members={tid: Tensor.from_xarray(data)}, stat={}, id=None)
    expected = xr.zeros_like(data)
    expected[{"x": slice(1, None)}] = 1
    op(sample)
    xr.testing.assert_allclose(expected, sample.members[tid].data)


def test_binarize2(tid: MemberId):
    from bioimageio.core.proc_ops import Binarize

    shape = (3, 32, 32)
    axes = ("channel", "y", "x")
    np_data = np.random.rand(*shape)
    data = xr.DataArray(np_data, dims=axes)

    threshold = 0.5
    exp = xr.DataArray(np_data > threshold, dims=axes)

    sample = Sample(members={tid: Tensor.from_xarray(data)}, stat={}, id=None)
    binarize = Binarize(tid, tid, threshold=threshold)
    binarize(sample)
    xr.testing.assert_allclose(exp, sample.members[tid].data)


def test_clip(tid: MemberId):
    from bioimageio.core.proc_ops import Clip

    op = Clip(tid, tid, min=3, max=5)
    data = xr.DataArray(np.arange(9).reshape(3, 3), dims=("x", "y"))
    sample = Sample(members={tid: Tensor.from_xarray(data)}, stat={}, id=None)

    expected = xr.DataArray(
        np.array([[3, 3, 3], [3, 4, 5], [5, 5, 5]]), dims=("x", "y")
    )
    op(sample)
    xr.testing.assert_equal(expected, sample.members[tid].data)


def test_combination_of_op_steps_with_dims_specified(tid: MemberId):
    from bioimageio.core.proc_ops import ZeroMeanUnitVariance

    data = xr.DataArray(np.arange(18).reshape((2, 3, 3)), dims=("channel", "x", "y"))
    sample = Sample(members={tid: Tensor.from_xarray(data)}, stat={}, id=None)
    op = ZeroMeanUnitVariance(
        tid,
        tid,
        SampleMean(
            member_id=tid,
            axes=(AxisId("x"), AxisId("y")),
        ),
        SampleStd(
            member_id=tid,
            axes=(AxisId("x"), AxisId("y")),
        ),
    )
    sample.stat = compute_measures(op.required_measures, [sample])

    expected = xr.DataArray(
        np.array(
            [
                [
                    [-1.54919274, -1.16189455, -0.77459637],
                    [-0.38729818, 0.0, 0.38729818],
                    [0.77459637, 1.16189455, 1.54919274],
                ],
                [
                    [-1.54919274, -1.16189455, -0.77459637],
                    [-0.38729818, 0.0, 0.38729818],
                    [0.77459637, 1.16189455, 1.54919274],
                ],
            ]
        ),
        dims=("channel", "x", "y"),
    )

    op(sample)
    xr.testing.assert_allclose(expected, sample.members[tid].data, rtol=1e-5, atol=1e-7)


@pytest.mark.parametrize(
    "axes",
    [
        None,
        tuple(map(AxisId, "cy")),
        tuple(map(AxisId, "cyx")),
        tuple(map(AxisId, "x")),
    ],
)
def test_scale_mean_variance(tid: MemberId, axes: Optional[Tuple[AxisId, ...]]):
    from bioimageio.core.proc_ops import ScaleMeanVariance

    shape = (3, 32, 46)
    ipt_axes = ("channel", "y", "x")
    np_data = np.random.rand(*shape)
    ipt_data = xr.DataArray(np_data, dims=ipt_axes)
    ref_data = xr.DataArray((np_data * 2) + 3, dims=ipt_axes)

    op = ScaleMeanVariance(tid, tid, reference_tensor=MemberId("ref_name"), axes=axes)
    sample = Sample(
        members={
            tid: Tensor.from_xarray(ipt_data),
            MemberId("ref_name"): Tensor.from_xarray(ref_data),
        },
        stat={},
        id=None,
    )
    sample.stat = compute_measures(op.required_measures, [sample])
    op(sample)
    xr.testing.assert_allclose(ref_data, sample.members[tid].data, rtol=1e-5, atol=1e-7)


@pytest.mark.parametrize(
    "axes_str",
    [None, "cy", "y", "yx"],
)
def test_scale_mean_variance_per_channel(tid: MemberId, axes_str: Optional[str]):
    from bioimageio.core.proc_ops import ScaleMeanVariance

    axes = None if axes_str is None else tuple(map(AxisId, axes_str))

    shape = (3, 32, 46)
    ipt_axes = ("channel", "y", "x")
    np_data = np.random.rand(*shape)
    ipt_data = xr.DataArray(np_data, dims=ipt_axes)

    # set different mean, std per channel
    np_ref_data = np.stack([d * i + i for i, d in enumerate(np_data, start=2)])
    ref_data = xr.DataArray(np_ref_data, dims=ipt_axes)

    op = ScaleMeanVariance(tid, tid, reference_tensor=MemberId("ref_name"), axes=axes)
    sample = Sample(
        members={
            tid: Tensor.from_xarray(ipt_data),
            MemberId("ref_name"): Tensor.from_xarray(ref_data),
        },
        stat={},
        id=None,
    )
    sample.stat = compute_measures(op.required_measures, [sample])
    op(sample)

    if axes is not None and AxisId("c") not in axes:
        # mean,std per channel should match exactly
        xr.testing.assert_allclose(
            ref_data, sample.members[tid].data, rtol=1e-5, atol=1e-7
        )
    else:
        # mean,std across channels should not match
        with pytest.raises(AssertionError):
            xr.testing.assert_allclose(
                ref_data, sample.members[tid].data, rtol=1e-5, atol=1e-7
            )


def test_scale_range(tid: MemberId):
    from bioimageio.core.proc_ops import ScaleRange

    op = ScaleRange(tid, tid)
    np_data = np.arange(9).reshape(3, 3).astype("float32")
    data = xr.DataArray(np_data, dims=("x", "y"))
    sample = Sample(members={tid: Tensor.from_xarray(data)}, stat={}, id=None)
    sample.stat = compute_measures(op.required_measures, [sample])

    eps = 1.0e-6
    mi, ma = np_data.min(), np_data.max()
    exp_data = (np_data - mi) / (ma - mi + eps)
    expected = xr.DataArray(exp_data, dims=("x", "y"))

    op(sample)
    # NOTE xarray.testing.assert_allclose compares irrelavant properties here and fails although the result is correct
    np.testing.assert_allclose(expected, sample.members[tid].data, rtol=1e-5, atol=1e-7)


def test_scale_range_axes(tid: MemberId):
    from bioimageio.core.proc_ops import ScaleRange

    eps = 1.0e-6

    lower_quantile = SampleQuantile(
        member_id=tid, q=0.1, axes=(AxisId("x"), AxisId("y"))
    )
    upper_quantile = SampleQuantile(
        member_id=tid, q=0.9, axes=(AxisId("x"), AxisId("y"))
    )
    op = ScaleRange(tid, tid, lower_quantile, upper_quantile, eps=eps)

    np_data = np.arange(18).reshape((2, 3, 3)).astype("float32")
    data = Tensor.from_xarray(xr.DataArray(np_data, dims=("channel", "x", "y")))
    sample = Sample(members={tid: data}, stat={}, id=None)

    p_low_direct = lower_quantile.compute(sample)
    p_up_direct = upper_quantile.compute(sample)

    p_low_expected = np.quantile(np_data, lower_quantile.q, axis=(1, 2), keepdims=True)
    p_up_expected = np.quantile(np_data, upper_quantile.q, axis=(1, 2), keepdims=True)

    np.testing.assert_allclose(p_low_expected.squeeze(), p_low_direct)
    np.testing.assert_allclose(p_up_expected.squeeze(), p_up_direct)

    sample.stat = compute_measures(op.required_measures, [sample])

    np.testing.assert_allclose(p_low_expected.squeeze(), sample.stat[lower_quantile])
    np.testing.assert_allclose(p_up_expected.squeeze(), sample.stat[upper_quantile])

    exp_data = (np_data - p_low_expected) / (p_up_expected - p_low_expected + eps)
    expected = xr.DataArray(exp_data, dims=("channel", "x", "y"))

    op(sample)
    # NOTE xarray.testing.assert_allclose compares irrelavant properties here and fails although the result is correct
    np.testing.assert_allclose(expected, sample.members[tid].data, rtol=1e-5, atol=1e-7)


def test_sigmoid(tid: MemberId):
    from bioimageio.core.proc_ops import Sigmoid

    shape = (3, 32, 32)
    axes = ("channel", "y", "x")
    np_data = np.random.rand(*shape)
    data = xr.DataArray(np_data, dims=axes)
    sample = Sample(members={tid: Tensor.from_xarray(data)}, stat={}, id=None)
    sigmoid = Sigmoid(tid, tid)
    sigmoid(sample)

    exp = xr.DataArray(1.0 / (1 + np.exp(-np_data)), dims=axes)
    xr.testing.assert_allclose(exp, sample.members[tid].data, rtol=1e-5, atol=1e-7)


def test_softmax(tid: MemberId):
    from bioimageio.core.proc_ops import Softmax

    shape = (3, 32, 32)
    axes = ("channel", "y", "x")
    np_data = np.random.rand(*shape)
    data = xr.DataArray(np_data, dims=axes)
    sample = Sample(members={tid: Tensor.from_xarray(data)}, stat={}, id=None)
    softmax = Softmax(tid, tid, axis=AxisId("channel"))
    softmax(sample)

    exp = xr.DataArray(
        np.exp(np_data - np.max(np_data, axis=0, keepdims=True))
        / np.sum(np.exp(np_data - np.max(np_data, axis=0, keepdims=True)), axis=0),
        dims=axes,
    )
    xr.testing.assert_allclose(exp, sample.members[tid].data, rtol=1e-5, atol=1e-7)


def test_softmax_with_scipy(tid: MemberId):
    from bioimageio.core.proc_ops import Softmax

    shape = (3, 32, 32)
    axes = ("channel", "y", "x")
    np_data = np.random.rand(*shape)
    data = xr.DataArray(np_data, dims=axes)
    sample = Sample(members={tid: Tensor.from_xarray(data)}, stat={}, id=None)
    softmax = Softmax(tid, tid, axis=AxisId("channel"))
    softmax(sample)

    exp = xr.DataArray(
        scipy.special.softmax(np_data, axis=0),
        dims=axes,
    )
    xr.testing.assert_allclose(exp, sample.members[tid].data, rtol=1e-5, atol=1e-7)



import os
import shutil
import sys
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Any, Union, no_type_check
from zipfile import ZipFile

import tensorflow  # pyright: ignore[reportMissingTypeStubs]

from bioimageio.spec._internal.version_type import Version
from bioimageio.spec.common import ZipPath
from bioimageio.spec.model.v0_5 import (
    InputTensorDescr,
    ModelDescr,
    OutputTensorDescr,
    TensorflowSavedModelBundleWeightsDescr,
)

from .. import __version__
from ..io import ensure_unzipped

try:
    # try to build the tf model with the keras import from tensorflow
    from tensorflow import keras  # type: ignore
except Exception:
    # if the above fails try to export with the standalone keras
    import keras  # pyright: ignore[reportMissingTypeStubs]


def convert(
    model_descr: ModelDescr, output_path: Path
) -> TensorflowSavedModelBundleWeightsDescr:
    """
    Convert model weights from the 'keras_hdf5' format to the 'tensorflow_saved_model_bundle' format.

    This method handles the conversion of Keras HDF5 model weights into a TensorFlow SavedModel bundle,
    which is the recommended format for deploying TensorFlow models. The method supports both TensorFlow 1.x
    and 2.x versions, with appropriate checks to ensure compatibility.

    Adapted from:
    https://github.com/deepimagej/pydeepimagej/blob/5aaf0e71f9b04df591d5ca596f0af633a7e024f5/pydeepimagej/yaml/create_config.py

    Args:
        model_descr:
            The bioimage.io model description containing the model's metadata and weights.
        output_path:
            Path with .zip suffix (.zip is appended otherwise) to which a zip archive
            with the TensorFlow SavedModel bundle will be saved.
    Raises:
        ValueError:
            - If the specified `output_path` already exists.
            - If the Keras HDF5 weights are missing in the model description.
        RuntimeError:
            If there is a mismatch between the TensorFlow version used by the model and the version installed.
        NotImplementedError:
            If the model has multiple inputs or outputs and TensorFlow 1.x is being used.

    Returns:
        A descriptor object containing information about the converted TensorFlow SavedModel bundle.
    """
    tf_major_ver = int(tensorflow.__version__.split(".")[0])

    if output_path.suffix != ".zip":
        output_path = output_path.with_suffix("")

    if output_path.exists():
        raise ValueError(f"The ouptut directory at {output_path} must not exist.")

    if model_descr.weights.keras_hdf5 is None:
        raise ValueError("Missing Keras Hdf5 weights to convert from.")

    weight_spec = model_descr.weights.keras_hdf5
    weight_reader = weight_spec.get_reader()

    if weight_spec.tensorflow_version:
        model_tf_major_ver = int(weight_spec.tensorflow_version.major)
        if model_tf_major_ver != tf_major_ver:
            raise RuntimeError(
                f"Tensorflow major versions of model {model_tf_major_ver} is not {tf_major_ver}"
            )

    td_kwargs: dict[str, Any] = (
        dict(ignore_cleanup_errors=True) if sys.version_info >= (3, 10) else {}
    )
    with TemporaryDirectory(**td_kwargs) as temp_dir:
        local_weights = ensure_unzipped(
            weight_reader, Path(temp_dir) / "bioimageio_unzipped_tf_weights"
        )
        if tf_major_ver == 1:
            if len(model_descr.inputs) != 1 or len(model_descr.outputs) != 1:
                raise NotImplementedError(
                    "Weight conversion for models with multiple inputs or outputs is not yet implemented."
                )

            input_name = str(
                d.id
                if isinstance((d := model_descr.inputs[0]), InputTensorDescr)
                else d.name
            )
            output_name = str(
                d.id
                if isinstance((d := model_descr.outputs[0]), OutputTensorDescr)
                else d.name
            )
            return _convert_tf1(
                ensure_unzipped(local_weights, Path("bioimageio_unzipped_tf_weights")),
                output_path,
                input_name,
                output_name,
            )
        else:
            return _convert_tf2(local_weights, output_path)


def _convert_tf2(
    keras_weight_path: Union[Path, ZipPath], output_path: Path
) -> TensorflowSavedModelBundleWeightsDescr:
    model = keras.models.load_model(keras_weight_path)  # type: ignore
    model.export(output_path)  # type: ignore

    output_path = _zip_model_bundle(output_path)
    print("TensorFlow model exported to", output_path)

    return TensorflowSavedModelBundleWeightsDescr(
        source=output_path.absolute(),
        parent="keras_hdf5",
        tensorflow_version=Version(tensorflow.__version__),
        comment=f"Converted with bioimageio.core {__version__}.",
    )


# adapted from
# https://github.com/deepimagej/pydeepimagej/blob/master/pydeepimagej/yaml/create_config.py#L236
def _convert_tf1(
    keras_weight_path: Path,
    output_path: Path,
    input_name: str,
    output_name: str,
) -> TensorflowSavedModelBundleWeightsDescr:
    @no_type_check
    def build_tf_model():
        keras_model = keras.models.load_model(keras_weight_path)
        builder = tensorflow.saved_model.builder.SavedModelBuilder(output_path)
        signature = tensorflow.saved_model.signature_def_utils.predict_signature_def(
            inputs={input_name: keras_model.input},
            outputs={output_name: keras_model.output},
        )

        signature_def_map = {
            tensorflow.saved_model.signature_constants.DEFAULT_SERVING_SIGNATURE_DEF_KEY: (
                signature
            )
        }

        builder.add_meta_graph_and_variables(
            keras.backend.get_session(),
            [tensorflow.saved_model.tag_constants.SERVING],
            signature_def_map=signature_def_map,
        )
        builder.save()

    build_tf_model()

    output_path = _zip_model_bundle(output_path)
    print("TensorFlow model exported to", output_path)

    return TensorflowSavedModelBundleWeightsDescr(
        source=output_path.absolute(),
        parent="keras_hdf5",
        tensorflow_version=Version(tensorflow.__version__),
        comment=f"Converted with bioimageio.core {__version__}.",
    )


def _zip_model_bundle(model_bundle_folder: Path):
    zipped_model_bundle = model_bundle_folder.with_suffix(".zip")

    with ZipFile(zipped_model_bundle, "w") as zip_obj:
        for root, _, files in os.walk(model_bundle_folder):
            for filename in files:
                src = os.path.join(root, filename)
                zip_obj.write(src, os.path.relpath(src, model_bundle_folder))

    try:
        shutil.rmtree(model_bundle_folder)
    except Exception:
        print("TensorFlow bundled model was not removed after compression")

    return zipped_model_bundle



import collections.abc
import warnings
import zipfile
from pathlib import Path, PurePosixPath
from shutil import copyfileobj
from typing import (
    Any,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    Union,
)

import h5py  # pyright: ignore[reportMissingTypeStubs]
from imageio.v3 import imread, imwrite  # type: ignore
from loguru import logger
from numpy.typing import NDArray
from pydantic import BaseModel, ConfigDict, TypeAdapter
from typing_extensions import assert_never

from bioimageio.spec._internal.io import get_reader, interprete_file_source
from bioimageio.spec._internal.type_guards import is_ndarray
from bioimageio.spec.common import (
    BytesReader,
    FileSource,
    HttpUrl,
    PermissiveFileSource,
    RelativeFilePath,
    ZipPath,
)
from bioimageio.spec.utils import download, load_array, save_array

from .axis import AxisLike
from .common import PerMember
from .sample import Sample
from .stat_measures import DatasetMeasure, MeasureValue
from .tensor import Tensor

DEFAULT_H5_DATASET_PATH = "data"


SUFFIXES_WITH_DATAPATH = (".h5", ".hdf", ".hdf5")


def load_image(
    source: Union[ZipPath, PermissiveFileSource], is_volume: Optional[bool] = None
) -> NDArray[Any]:
    """load a single image as numpy array

    Args:
        source: image source
        is_volume: deprecated
    """
    if is_volume is not None:
        warnings.warn("**is_volume** is deprecated and will be removed soon.")

    if isinstance(source, ZipPath):
        parsed_source = source
    else:
        parsed_source = interprete_file_source(source)

    if isinstance(parsed_source, RelativeFilePath):
        src = parsed_source.absolute()
    else:
        src = parsed_source

    if isinstance(src, Path):
        file_source, suffix, subpath = _split_dataset_path(src)
    elif isinstance(src, HttpUrl):
        file_source, suffix, subpath = _split_dataset_path(src)
    elif isinstance(src, ZipPath):
        file_source, suffix, subpath = _split_dataset_path(src)
    else:
        assert_never(src)

    if suffix == ".npy":
        if subpath is not None:
            logger.warning(
                "Unexpected subpath {} for .npy source {}", subpath, file_source
            )

        image = load_array(file_source)
    elif suffix in SUFFIXES_WITH_DATAPATH:
        if subpath is None:
            dataset_path = DEFAULT_H5_DATASET_PATH
        else:
            dataset_path = str(subpath)

        reader = download(file_source)

        with h5py.File(reader, "r") as f:
            h5_dataset = f.get(  # pyright: ignore[reportUnknownVariableType]
                dataset_path
            )
            if not isinstance(h5_dataset, h5py.Dataset):
                raise ValueError(
                    f"{file_source} did not load as {h5py.Dataset}, but has type "
                    + str(
                        type(h5_dataset)  # pyright: ignore[reportUnknownArgumentType]
                    )
                )
            image: NDArray[Any]
            image = h5_dataset[:]  # pyright: ignore[reportUnknownVariableType]
    else:
        reader = download(file_source)
        image = imread(  # pyright: ignore[reportUnknownVariableType]
            reader.read(), extension=suffix
        )

    assert is_ndarray(image)
    return image


def load_tensor(
    path: Union[ZipPath, Path, str], axes: Optional[Sequence[AxisLike]] = None
) -> Tensor:
    # TODO: load axis meta data
    array = load_image(path)

    return Tensor.from_numpy(array, dims=axes)


_SourceT = TypeVar("_SourceT", Path, HttpUrl, ZipPath)

Suffix = str


def _split_dataset_path(
    source: _SourceT,
) -> Tuple[_SourceT, Suffix, Optional[PurePosixPath]]:
    """Split off subpath (e.g. internal  h5 dataset path)
    from a file path following a file extension.

    Examples:
        >>> _split_dataset_path(Path("my_file.h5/dataset"))
        (...Path('my_file.h5'), '.h5', PurePosixPath('dataset'))

        >>> _split_dataset_path(Path("my_plain_file"))
        (...Path('my_plain_file'), '', None)

    """
    if isinstance(source, RelativeFilePath):
        src = source.absolute()
    else:
        src = source

    del source

    def separate_pure_path(path: PurePosixPath):
        for p in path.parents:
            if p.suffix in SUFFIXES_WITH_DATAPATH:
                return p, p.suffix, PurePosixPath(path.relative_to(p))

        return path, path.suffix, None

    if isinstance(src, HttpUrl):
        file_path, suffix, data_path = separate_pure_path(PurePosixPath(src.path or ""))

        if data_path is None:
            return src, suffix, None

        return (
            HttpUrl(str(file_path).replace(f"/{data_path}", "")),
            suffix,
            data_path,
        )

    if isinstance(src, ZipPath):
        file_path, suffix, data_path = separate_pure_path(PurePosixPath(str(src)))

        if data_path is None:
            return src, suffix, None

        return (
            ZipPath(str(file_path).replace(f"/{data_path}", "")),
            suffix,
            data_path,
        )

    file_path, suffix, data_path = separate_pure_path(PurePosixPath(src))
    return Path(file_path), suffix, data_path


def save_tensor(path: Union[Path, str], tensor: Tensor) -> None:
    # TODO: save axis meta data

    data: NDArray[Any] = (  # pyright: ignore[reportUnknownVariableType]
        tensor.data.to_numpy()
    )
    assert is_ndarray(data)
    file_path, suffix, subpath = _split_dataset_path(Path(path))
    if not suffix:
        raise ValueError(f"No suffix (needed to decide file format) found in {path}")

    file_path.parent.mkdir(exist_ok=True, parents=True)
    if file_path.suffix == ".npy":
        if subpath is not None:
            raise ValueError(f"Unexpected subpath {subpath} found in .npy path {path}")
        save_array(file_path, data)
    elif suffix in (".h5", ".hdf", ".hdf5"):
        if subpath is None:
            dataset_path = DEFAULT_H5_DATASET_PATH
        else:
            dataset_path = str(subpath)

        with h5py.File(file_path, "a") as f:
            if dataset_path in f:
                del f[dataset_path]

            _ = f.create_dataset(dataset_path, data=data, chunks=True)
    else:
        # if singleton_axes := [a for a, s in tensor.tagged_shape.items() if s == 1]:
        #     tensor = tensor[{a: 0 for a in singleton_axes}]
        #     singleton_axes_msg = f"(without singleton axes {singleton_axes}) "
        # else:
        singleton_axes_msg = ""

        logger.debug(
            "writing tensor {} {}to {}",
            dict(tensor.tagged_shape),
            singleton_axes_msg,
            path,
        )
        imwrite(path, data)


def save_sample(
    path: Union[Path, str, PerMember[Union[Path, str]]], sample: Sample
) -> None:
    """Save a **sample** to a **path** pattern
    or all sample members in the **path** mapping.

    If **path** is a pathlib.Path or a string and the **sample** has multiple members,
    **path** it must contain `{member_id}` (or `{input_id}` or `{output_id}`).

    (Each) **path** may contain `{sample_id}` to be formatted with the **sample** object.
    """
    if not isinstance(path, collections.abc.Mapping):
        if len(sample.members) < 2 or any(
            m in str(path) for m in ("{member_id}", "{input_id}", "{output_id}")
        ):
            path = {m: path for m in sample.members}
        else:
            raise ValueError(
                f"path {path} must contain '{{member_id}}' for sample with multiple members {list(sample.members)}."
            )

    for m, p in path.items():
        t = sample.members[m]
        p_formatted = Path(
            str(p).format(sample_id=sample.id, member_id=m, input_id=m, output_id=m)
        )
        save_tensor(p_formatted, t)


class _SerializedDatasetStatsEntry(
    BaseModel, frozen=True, arbitrary_types_allowed=True
):
    measure: DatasetMeasure
    value: MeasureValue


_stat_adapter = TypeAdapter(
    Sequence[_SerializedDatasetStatsEntry],
    config=ConfigDict(arbitrary_types_allowed=True),
)


def save_dataset_stat(stat: Mapping[DatasetMeasure, MeasureValue], path: Path):
    serializable = [
        _SerializedDatasetStatsEntry(measure=k, value=v) for k, v in stat.items()
    ]
    _ = path.write_bytes(_stat_adapter.dump_json(serializable))


def load_dataset_stat(path: Path):
    seq = _stat_adapter.validate_json(path.read_bytes())
    return {e.measure: e.value for e in seq}


def ensure_unzipped(
    source: Union[PermissiveFileSource, ZipPath, BytesReader], folder: Path
):
    """unzip a (downloaded) **source** to a file in **folder** if source is a zip archive
    otherwise copy **source** to a file in **folder**."""
    if isinstance(source, BytesReader):
        weights_reader = source
    else:
        weights_reader = get_reader(source)

    out_path = folder / (
        weights_reader.original_file_name or f"file{weights_reader.suffix}"
    )

    if zipfile.is_zipfile(weights_reader):
        out_path = out_path.with_name(out_path.name + ".unzipped")
        out_path.parent.mkdir(exist_ok=True, parents=True)
        # source itself is a zipfile
        with zipfile.ZipFile(weights_reader, "r") as f:
            f.extractall(out_path)

    else:
        out_path.parent.mkdir(exist_ok=True, parents=True)
        with out_path.open("wb") as f:
            copyfileobj(weights_reader, f)

    return out_path


def get_suffix(source: Union[ZipPath, FileSource]) -> str:
    if isinstance(source, Path):
        return source.suffix
    elif isinstance(source, ZipPath):
        return source.suffix
    if isinstance(source, RelativeFilePath):
        return source.path.suffix
    elif isinstance(source, ZipPath):
        return source.suffix
    elif isinstance(source, HttpUrl):
        if source.path is None:
            return ""
        else:
            return PurePosixPath(source.path).suffix
    else:
        assert_never(source)



# pyright: reportUnknownVariableType=false
import warnings
from typing import Any, List, Optional, Sequence, Union

import onnxruntime as rt  # pyright: ignore[reportMissingTypeStubs]
from numpy.typing import NDArray

from bioimageio.spec.model import v0_4, v0_5

from ..model_adapters import ModelAdapter
from ..utils._type_guards import is_list, is_tuple


class ONNXModelAdapter(ModelAdapter):
    def __init__(
        self,
        *,
        model_description: Union[v0_4.ModelDescr, v0_5.ModelDescr],
        devices: Optional[Sequence[str]] = None,
    ):
        super().__init__(model_description=model_description)

        if model_description.weights.onnx is None:
            raise ValueError("No ONNX weights specified for {model_description.name}")

        reader = model_description.weights.onnx.get_reader()
        self._session = rt.InferenceSession(reader.read())
        onnx_inputs = self._session.get_inputs()
        self._input_names: List[str] = [ipt.name for ipt in onnx_inputs]

        if devices is not None:
            warnings.warn(
                f"Device management is not implemented for onnx yet, ignoring the devices {devices}"
            )

    def _forward_impl(
        self, input_arrays: Sequence[Optional[NDArray[Any]]]
    ) -> List[Optional[NDArray[Any]]]:
        result: Any = self._session.run(
            None, dict(zip(self._input_names, input_arrays))
        )
        if is_list(result) or is_tuple(result):
            result_seq = list(result)
        else:
            result_seq = [result]

        return result_seq

    def unload(self) -> None:
        warnings.warn(
            "Device management is not implemented for onnx yet, cannot unload model"
        )



import os
from itertools import chain
from pathlib import Path
from typing import Iterable, Mapping, Tuple

import pytest
from bioimageio.spec import InvalidDescr, settings
from bioimageio.spec.common import Sha256
from pydantic import HttpUrl

from tests.utils import ParameterSet, expensive_test

TEST_RDF_SOURCES: Mapping[str, Tuple[HttpUrl, Sha256]] = {
    # "affable-shark": (  # TODO: enable when updated with fixed torchscript and onnx weights
    #     HttpUrl(
    #         "https://hypha.aicell.io/bioimage-io/artifacts/affable-shark/files/rdf.yaml?version=v0"
    #     ),
    #     Sha256("b74944b4949591d3eaf231cf9ab259f91dec679863020178e6c3ddadd52a019c"),
    # ),
    "ambitious-sloth": (
        HttpUrl(
            "https://hypha.aicell.io/bioimage-io/artifacts/ambitious-sloth/files/rdf.yaml?version=v0"
        ),
        Sha256("caf162e847a0812fb7704e7848b1ee68f46383278d8b74493553fe96750d1e39"),
    ),
}


def yield_bioimageio_yaml_urls() -> Iterable[ParameterSet]:
    for key, (descr_url, sha) in TEST_RDF_SOURCES.items():
        yield pytest.param(descr_url, sha, key, id=key)


def get_directory_size(path: Path):
    total_size = 0
    for dirpath, _, filenames in os.walk(path):
        for f in filenames:
            fp = os.path.join(dirpath, f)
            # skip if it is symbolic link
            if not os.path.islink(fp):
                total_size += os.path.getsize(fp)

    return total_size


@pytest.mark.parametrize("descr_url,sha,key", list(yield_bioimageio_yaml_urls()))
def test_rdf_format_to_populate_cache(
    descr_url: HttpUrl,
    sha: Sha256,
    key: str,
):
    """this test is redundant if `test_rdf` runs, but is used in the CI to populate the cache"""
    from bioimageio.spec._internal.gh_utils import set_github_warning

    if os.environ.get("BIOIMAGEIO_POPULATE_CACHE") != "1":
        pytest.skip("BIOIMAGEIO_POPULATE_CACHE != 1")

    if (cache_size := get_directory_size(settings.cache_path)) > 7e9:
        msg = f"Reached 7GB cache size limit ({cache_size / 1e9:.2f} GB)"
        set_github_warning("Reached cache size limit", msg)
        pytest.skip(msg)

    from bioimageio.core import load_description

    _ = load_description(descr_url, sha256=sha, perform_io_checks=True)


@expensive_test
@pytest.mark.parametrize("descr_url,sha,key", list(yield_bioimageio_yaml_urls()))
def test_rdf(
    descr_url: HttpUrl,
    sha: Sha256,
    key: str,
):
    from bioimageio.spec import get_conda_env
    from bioimageio.spec.model import ModelDescr

    from bioimageio.core import load_description, load_description_and_test

    descr = load_description(
        descr_url, sha256=sha, format_version="latest", perform_io_checks=True
    )
    assert not isinstance(descr, InvalidDescr), descr.validation_summary.display() or [
        e.msg for e in descr.validation_summary.errors
    ]

    if (
        isinstance(descr, ModelDescr)
        and descr.weights.pytorch_state_dict is not None
        and descr.weights.pytorch_state_dict.dependencies is not None
    ):
        conda_env = get_conda_env(entry=descr.weights.pytorch_state_dict)

        def depends_on(dep: str) -> bool:
            return any(
                chain(
                    (d.startswith(dep) for d in conda_env.get_pip_deps()),
                    (cd for cd in conda_env.dependencies if isinstance(cd, str)),
                )
            )

        for skip_if_depends_on in (
            "biapy",
            "git+https://github.com/CAREamics/careamics.git",
            "careamics",
            "inferno",
            "plantseg",
        ):
            if depends_on(skip_if_depends_on):
                pytest.skip(f"requires {skip_if_depends_on}")

    descr = load_description_and_test(
        descr,
        format_version="latest",
        sha256=sha,
        stop_early=True,
    )

    assert not isinstance(descr, InvalidDescr), descr.validation_summary.display() or [
        e.msg for e in descr.validation_summary.errors
    ]
    assert descr.validation_summary.status == "passed", (
        descr.validation_summary.display()
        or [e.msg for e in descr.validation_summary.errors]
    )



"""These functions are used in the bioimageio command line interface
defined in `bioimageio.core.cli`."""

from pathlib import Path
from typing import Optional, Sequence, Union

from typing_extensions import Literal

from bioimageio.spec import (
    InvalidDescr,
    ResourceDescr,
    save_bioimageio_package,
    save_bioimageio_package_as_folder,
)
from bioimageio.spec._internal.types import FormatVersionPlaceholder

from ._resource_tests import test_description

# unfortunately this does not work with py3.9 and pydantic 2.11
# from bioimageio.core.common import SupportedWeightsFormat
# WeightFormatArgAll = Literal[SupportedWeightsFormat, "all"]
# WeightFormatArgAny = Literal[SupportedWeightsFormat, "any"]
# so we write out the literal explicitly
WeightFormatArgAll = Literal[
    "keras_hdf5",
    "onnx",
    "pytorch_state_dict",
    "tensorflow_saved_model_bundle",
    "torchscript",
    "all",
]
WeightFormatArgAny = Literal[
    "keras_hdf5",
    "onnx",
    "pytorch_state_dict",
    "tensorflow_saved_model_bundle",
    "torchscript",
    "any",
]


def test(
    descr: Union[ResourceDescr, InvalidDescr],
    *,
    weight_format: WeightFormatArgAll = "all",
    devices: Optional[Union[str, Sequence[str]]] = None,
    summary: Union[
        Literal["display"], Path, Sequence[Union[Literal["display"], Path]]
    ] = "display",
    runtime_env: Union[
        Literal["currently-active", "as-described"], Path
    ] = "currently-active",
    determinism: Literal["seed_only", "full"] = "seed_only",
    format_version: Union[FormatVersionPlaceholder, str] = "discover",
) -> int:
    """Test a bioimageio resource.

    Arguments as described in `bioimageio.core.cli.TestCmd`
    """
    if isinstance(descr, InvalidDescr):
        test_summary = descr.validation_summary
    else:
        test_summary = test_description(
            descr,
            format_version=format_version,
            weight_format=None if weight_format == "all" else weight_format,
            devices=[devices] if isinstance(devices, str) else devices,
            runtime_env=runtime_env,
            determinism=determinism,
        )

    _ = test_summary.log(summary)
    return 0 if test_summary.status == "passed" else 1


def validate_format(
    descr: Union[ResourceDescr, InvalidDescr],
    summary: Union[Path, Sequence[Path]] = (),
):
    """DEPRECATED; Access the existing `validation_summary` attribute instead.
    validate the meta data format of a bioimageio resource

    Args:
        descr: a bioimageio resource description
    """
    _ = descr.validation_summary.save(summary)
    return 0 if descr.validation_summary.status in ("valid-format", "passed") else 1


# TODO: absorb into `save_bioimageio_package`
def package(
    descr: ResourceDescr,
    path: Path,
    *,
    weight_format: WeightFormatArgAll = "all",
):
    """Save a resource's metadata with its associated files.

    Note: If `path` does not have a `.zip` suffix this command will save the
          package as an unzipped folder instead.

    Args:
        descr: a bioimageio resource description
        path: output path
        weight-format: include only this single weight-format (if not 'all').
    """
    if isinstance(descr, InvalidDescr):
        logged = descr.validation_summary.save()
        msg = f"Invalid {descr.type} description."
        if logged:
            msg += f" Details saved to {logged}."

        raise ValueError(msg)

    if weight_format == "all":
        weights_priority_order = None
    else:
        weights_priority_order = (weight_format,)

    if path.suffix == ".zip":
        _ = save_bioimageio_package(
            descr,
            output_path=path,
            weights_priority_order=weights_priority_order,
        )
    else:
        _ = save_bioimageio_package_as_folder(
            descr,
            output_path=path,
            weights_priority_order=weights_priority_order,
        )
    return 0



import json
import subprocess
from typing import Optional

import pytest
from bioimageio.spec._internal.gh_utils import set_github_warning
from packaging.version import Version


def test_bioimageio_spec_version(conda_cmd: Optional[str]):
    if conda_cmd is None:
        pytest.skip("requires conda")

    from importlib.metadata import metadata

    # get latest released bioimageio.spec version
    conda_search = subprocess.run(
        f"{conda_cmd} search --json -f conda-forge::bioimageio.spec>=0.5.3.2".split(),
        encoding="utf-8",
        capture_output=True,
        check=True,
    )
    result = json.loads(conda_search.stdout)
    latest_spec = max(result["bioimageio.spec"], key=lambda entry: entry["timestamp"])
    released = Version(latest_spec["version"])

    # get currently pinned bioimageio.spec version
    meta = metadata("bioimageio.core")
    req = meta["Requires-Dist"]
    valid_starts = ("bioimageio.spec ==", "bioimageio.spec==")
    for start in valid_starts:
        if req.startswith(start):
            spec_ver = req[len(start) :]
            break
    else:
        raise ValueError(
            f"Expected bioimageio.sepc pin to start with any of {valid_starts}"
        )

    assert spec_ver.count(".") == 3
    pinned = Version(spec_ver)
    if pinned != released:
        set_github_warning(
            "spec pin mismatch",
            f"bioimageio.spec pinned to {pinned}, while latest version on conda-forge is {released}",
        )



from itertools import product
from typing import Optional, Tuple

import numpy as np
import pytest
import xarray as xr

from bioimageio.core import stat_measures
from bioimageio.core.axis import AxisId
from bioimageio.core.common import MemberId
from bioimageio.core.sample import Sample
from bioimageio.core.stat_calculators import (
    SamplePercentilesCalculator,
    get_measure_calculators,
)
from bioimageio.core.stat_measures import SampleQuantile
from bioimageio.core.tensor import Tensor


@pytest.mark.parametrize(
    "name,axes",
    product(
        ["mean", "var", "std"],
        [None, (AxisId("c"),), (AxisId("x"), AxisId("y"))],
    ),
)
def test_individual_normal_measure(
    name: str,
    axes: Optional[Tuple[AxisId, AxisId]],
):
    data_id = MemberId("test_data")
    measure = getattr(stat_measures, "Sample" + name.title())(
        axes=axes, member_id=data_id
    )
    data = Tensor(
        np.random.random((5, 6, 3)), dims=(AxisId("x"), AxisId("y"), AxisId("c"))
    )

    expected = getattr(data, name)(dim=axes)
    sample = Sample(members={data_id: data}, stat={}, id=None)
    actual = measure.compute(sample)
    xr.testing.assert_allclose(expected.data, actual.data)


@pytest.mark.parametrize("axes", [None, (AxisId("x"), AxisId("y"))])
def test_individual_percentile_measure(axes: Optional[Tuple[AxisId, ...]]):
    qs = [0, 0.1, 0.5, 1.0]
    tid = MemberId("tensor")

    measures = [SampleQuantile(member_id=tid, axes=axes, q=q) for q in qs]
    calcs, _ = get_measure_calculators(measures)
    assert len(calcs) == 1
    calc = calcs[0]
    assert isinstance(calc, SamplePercentilesCalculator)

    data = Tensor(
        np.random.random((5, 6, 3)), dims=(AxisId("x"), AxisId("y"), AxisId("c"))
    )
    actual = calc.compute(Sample(members={tid: data}, stat={}, id=None))
    for m in measures:
        expected = data.quantile(q=m.q, dim=m.axes)
        actual_data = actual[m]
        if isinstance(actual_data, Tensor):
            actual_data = actual_data.data

        xr.testing.assert_allclose(expected.data, actual_data)



import os
import shutil
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Any, Optional, Sequence, Union

import h5py  # pyright: ignore[reportMissingTypeStubs]
from keras.src.legacy.saving import (  # pyright: ignore[reportMissingTypeStubs]
    legacy_h5_format,
)
from loguru import logger
from numpy.typing import NDArray

from bioimageio.spec.model import v0_4, v0_5
from bioimageio.spec.model.v0_5 import Version

from .._settings import settings
from ..digest_spec import get_axes_infos
from ..utils._type_guards import is_list, is_tuple
from ._model_adapter import ModelAdapter

os.environ["KERAS_BACKEND"] = settings.keras_backend


# by default, we use the keras integrated with tensorflow
# TODO: check if we should prefer keras
try:
    import tensorflow as tf  # pyright: ignore[reportMissingTypeStubs]
    from tensorflow import (  # pyright: ignore[reportMissingTypeStubs]
        keras,  # pyright: ignore[reportUnknownVariableType,reportAttributeAccessIssue]
    )

    tf_version = Version(tf.__version__)
except Exception:
    import keras  # pyright: ignore[reportMissingTypeStubs]

    tf_version = None


class KerasModelAdapter(ModelAdapter):
    def __init__(
        self,
        *,
        model_description: Union[v0_4.ModelDescr, v0_5.ModelDescr],
        devices: Optional[Sequence[str]] = None,
    ) -> None:
        super().__init__(model_description=model_description)
        if model_description.weights.keras_hdf5 is None:
            raise ValueError("model has not keras_hdf5 weights specified")
        model_tf_version = model_description.weights.keras_hdf5.tensorflow_version

        if tf_version is None or model_tf_version is None:
            logger.warning("Could not check tensorflow versions.")
        elif model_tf_version > tf_version:
            logger.warning(
                "The model specifies a newer tensorflow version than installed: {} > {}.",
                model_tf_version,
                tf_version,
            )
        elif (model_tf_version.major, model_tf_version.minor) != (
            tf_version.major,
            tf_version.minor,
        ):
            logger.warning(
                "Model tensorflow version {} does not match {}.",
                model_tf_version,
                tf_version,
            )

        # TODO keras device management
        if devices is not None:
            logger.warning(
                "Device management is not implemented for keras yet, ignoring the devices {}",
                devices,
            )

        weight_reader = model_description.weights.keras_hdf5.get_reader()
        if weight_reader.suffix in (".h5", "hdf5"):
            h5_file = h5py.File(weight_reader, mode="r")
            self._network = legacy_h5_format.load_model_from_hdf5(h5_file)
        else:
            with TemporaryDirectory() as temp_dir:
                temp_path = Path(temp_dir) / weight_reader.original_file_name
                with temp_path.open("wb") as f:
                    shutil.copyfileobj(weight_reader, f)

                self._network = keras.models.load_model(temp_path)

        self._output_axes = [
            tuple(a.id for a in get_axes_infos(out))
            for out in model_description.outputs
        ]

    def _forward_impl(  # pyright: ignore[reportUnknownParameterType]
        self, input_arrays: Sequence[Optional[NDArray[Any]]]
    ):
        network_output = self._network.predict(*input_arrays)  # type: ignore
        if is_list(network_output) or is_tuple(network_output):
            return network_output
        else:
            return [network_output]  # pyright: ignore[reportUnknownVariableType]

    def unload(self) -> None:
        logger.warning(
            "Device management is not implemented for keras yet, cannot unload model"
        )



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Collection, Union

from .sample import Sample, SampleBlock, SampleBlockWithOrigin
from .stat_measures import Measure


@dataclass
class Operator(ABC):
    @abstractmethod
    def __call__(self, sample: Union[Sample, SampleBlockWithOrigin]) -> None: ...

    @property
    @abstractmethod
    def required_measures(self) -> Collection[Measure]: ...


@dataclass
class BlockedOperator(Operator, ABC):
    @abstractmethod
    def __call__(
        self, sample: Union[Sample, SampleBlock, SampleBlockWithOrigin]
    ) -> None: ...

    @property
    @abstractmethod
    def required_measures(self) -> Collection[Measure]: ...
